@@ -24,15 +24,20 @@
  * SUCH DAMAGE.
  */
 
-#include <stdlib.h>
-#include <unistd.h>
-#include <stdio.h>
+#ifdef LIBUSB_GLOBAL_INCLUDE_FILE
+#include LIBUSB_GLOBAL_INCLUDE_FILE
+#else
+#include <errno.h>
 #include <poll.h>
 #include <pthread.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <time.h>
-#include <errno.h>
+#include <unistd.h>
 #include <sys/queue.h>
 #include <sys/endian.h>
+#endif
 
 #define	libusb_device_handle libusb20_device
 
@@ -141,7 +146,7 @@ libusb10_handle_events_sub(struct libusb_context *ctx, struct timeval *tv)
 		err = LIBUSB_ERROR_IO;
 
 	if (err < 1) {
-		for (i = 0; i != nfds; i++) {
+		for (i = 0; i != (int)nfds; i++) {
 			if (ppdev[i] != NULL) {
 				CTX_UNLOCK(ctx);
 				libusb_unref_device(libusb_get_device(ppdev[i]));
@@ -150,7 +155,7 @@ libusb10_handle_events_sub(struct libusb_context *ctx, struct timeval *tv)
 		}
 		goto do_done;
 	}
-	for (i = 0; i != nfds; i++) {
+	for (i = 0; i != (int)nfds; i++) {
 		if (ppdev[i] != NULL) {
 			dev = libusb_get_device(ppdev[i]);
 
@@ -187,6 +192,8 @@ do_done:
 	/* Do all done callbacks */
 
 	while ((sxfer = TAILQ_FIRST(&ctx->tr_done))) {
+		uint8_t flags;
+
 		TAILQ_REMOVE(&ctx->tr_done, sxfer, entry);
 		sxfer->entry.tqe_prev = NULL;
 
@@ -197,13 +204,14 @@ do_done:
 		uxfer = (struct libusb_transfer *)(
 		    ((uint8_t *)sxfer) + sizeof(*sxfer));
 
+		/* Allow the callback to free the transfer itself. */
+		flags = uxfer->flags;
+
 		if (uxfer->callback != NULL)
 			(uxfer->callback) (uxfer);
 
-		if (uxfer->flags & LIBUSB_TRANSFER_FREE_BUFFER)
-			free(uxfer->buffer);
-
-		if (uxfer->flags & LIBUSB_TRANSFER_FREE_TRANSFER)
+		/* Check if the USB transfer should be automatically freed. */
+		if (flags & LIBUSB_TRANSFER_FREE_TRANSFER)
 			libusb_free_transfer(uxfer);
 
 		CTX_LOCK(ctx);
@@ -304,12 +312,16 @@ libusb_wait_for_event(libusb_context *ctx, struct timeval *tv)
 		    &ctx->ctx_lock);
 		return (0);
 	}
-	err = clock_gettime(CLOCK_REALTIME, &ts);
+	err = clock_gettime(CLOCK_MONOTONIC, &ts);
 	if (err < 0)
 		return (LIBUSB_ERROR_OTHER);
 
-	ts.tv_sec = tv->tv_sec;
-	ts.tv_nsec = tv->tv_usec * 1000;
+	/*
+	 * The "tv" arguments points to a relative time structure and
+	 * not an absolute time structure.
+	 */
+	ts.tv_sec += tv->tv_sec;
+	ts.tv_nsec += tv->tv_usec * 1000;
 	if (ts.tv_nsec >= 1000000000) {
 		ts.tv_nsec -= 1000000000;
 		ts.tv_sec++;
@@ -385,7 +397,7 @@ libusb_set_pollfd_notifiers(libusb_context *ctx,
 	ctx->fd_cb_user_data = user_data;
 }
 
-struct libusb_pollfd **
+const struct libusb_pollfd **
 libusb_get_pollfds(libusb_context *ctx)
 {
 	struct libusb_super_pollfd *pollfd;
@@ -411,7 +423,7 @@ libusb_get_pollfds(libusb_context *ctx)
 
 done:
 	CTX_UNLOCK(ctx);
-	return (ret);
+	return ((const struct libusb_pollfd **)ret);
 }
 
 
@@ -478,7 +490,7 @@ libusb10_do_transfer(libusb_device_handle *devh,
 {
 	libusb_context *ctx;
 	struct libusb_transfer *xfer;
-	volatile int complet;
+	int done;
 	int ret;
 
 	if (devh == NULL)
@@ -499,15 +511,15 @@ libusb10_do_transfer(libusb_device_handle *devh,
 	xfer->timeout = timeout;
 	xfer->buffer = data;
 	xfer->length = length;
-	xfer->user_data = (void *)&complet;
+	xfer->user_data = (void *)&done;
 	xfer->callback = libusb10_do_transfer_cb;
-	complet = 0;
+	done = 0;
 
 	if ((ret = libusb_submit_transfer(xfer)) < 0) {
 		libusb_free_transfer(xfer);
 		return (ret);
 	}
-	while (complet == 0) {
+	while (done == 0) {
 		if ((ret = libusb_handle_events(ctx)) < 0) {
 			libusb_cancel_transfer(xfer);
 			usleep(1000);	/* nice it */
@@ -578,7 +590,7 @@ libusb_interrupt_transfer(libusb_device_handle *devh,
 }
 
 uint8_t *
-libusb_get_iso_packet_buffer(struct libusb_transfer *transfer, uint32_t index)
+libusb_get_iso_packet_buffer(struct libusb_transfer *transfer, uint32_t off)
 {
 	uint8_t *ptr;
 	uint32_t n;
@@ -586,35 +598,35 @@ libusb_get_iso_packet_buffer(struct libusb_transfer *transfer, uint32_t index)
 	if (transfer->num_iso_packets < 0)
 		return (NULL);
 
-	if (index >= (uint32_t)transfer->num_iso_packets)
+	if (off >= (uint32_t)transfer->num_iso_packets)
 		return (NULL);
 
 	ptr = transfer->buffer;
 	if (ptr == NULL)
 		return (NULL);
 
-	for (n = 0; n != index; n++) {
+	for (n = 0; n != off; n++) {
 		ptr += transfer->iso_packet_desc[n].length;
 	}
 	return (ptr);
 }
 
 uint8_t *
-libusb_get_iso_packet_buffer_simple(struct libusb_transfer *transfer, uint32_t index)
+libusb_get_iso_packet_buffer_simple(struct libusb_transfer *transfer, uint32_t off)
 {
 	uint8_t *ptr;
 
 	if (transfer->num_iso_packets < 0)
 		return (NULL);
 
-	if (index >= (uint32_t)transfer->num_iso_packets)
+	if (off >= (uint32_t)transfer->num_iso_packets)
 		return (NULL);
 
 	ptr = transfer->buffer;
 	if (ptr == NULL)
 		return (NULL);
 
-	ptr += transfer->iso_packet_desc[0].length * index;
+	ptr += transfer->iso_packet_desc[0].length * off;
 
 	return (ptr);
 }
